import os

from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.primitives import padding


class VulnerableCryptor:
    def __init__(self):
        self.block_size = 16
        self.key = os.urandom(16)
        self.padding = padding.PKCS7(128)

    def encrypt(self, plaintext):
        padder = self.padding.padder()
        plaintext = padder.update(plaintext) + padder.finalize()

        iv = os.urandom(self.block_size)
        encryptor = Cipher(algorithms.AES(self.key), modes.CBC(iv)).encryptor()
        ciphertext = encryptor.update(plaintext) + encryptor.finalize()

        return iv + ciphertext

    def decrypt(self, ciphertext):
        iv = ciphertext[:self.block_size]
        ciphertext = ciphertext[self.block_size:]

        decryptor = Cipher(algorithms.AES(self.key), modes.CBC(iv)).decryptor()
        plaintext = decryptor.update(ciphertext) + decryptor.finalize()

        unpadder = self.padding.unpadder()
        unpadded = unpadder.update(plaintext) + unpadder.finalize()

        return unpadded

    def oracle(self, data):
        try:
            self.decrypt(data)
            return True
        except Exception:
            return False



# traceback_exception_init() adapted from trio
#
# _ExceptionPrintContext and traceback_exception_format() copied from the standard
# library
from __future__ import annotations

import collections.abc
import sys
import textwrap
import traceback
from functools import singledispatch
from types import TracebackType
from typing import Any, List, Optional

from ._exceptions import BaseExceptionGroup

max_group_width = 15
max_group_depth = 10
_cause_message = (
    "\nThe above exception was the direct cause of the following exception:\n\n"
)

_context_message = (
    "\nDuring handling of the above exception, another exception occurred:\n\n"
)


def _format_final_exc_line(etype, value):
    valuestr = _safe_string(value, "exception")
    if value is None or not valuestr:
        line = f"{etype}\n"
    else:
        line = f"{etype}: {valuestr}\n"

    return line


def _safe_string(value, what, func=str):
    try:
        return func(value)
    except BaseException:
        return f"<{what} {func.__name__}() failed>"


class _ExceptionPrintContext:
    def __init__(self):
        self.seen = set()
        self.exception_group_depth = 0
        self.need_close = False

    def indent(self):
        return " " * (2 * self.exception_group_depth)

    def emit(self, text_gen, margin_char=None):
        if margin_char is None:
            margin_char = "|"
        indent_str = self.indent()
        if self.exception_group_depth:
            indent_str += margin_char + " "

        if isinstance(text_gen, str):
            yield textwrap.indent(text_gen, indent_str, lambda line: True)
        else:
            for text in text_gen:
                yield textwrap.indent(text, indent_str, lambda line: True)


def exceptiongroup_excepthook(
    etype: type[BaseException], value: BaseException, tb: TracebackType | None
) -> None:
    sys.stderr.write("".join(traceback.format_exception(etype, value, tb)))


class PatchedTracebackException(traceback.TracebackException):
    def __init__(
        self,
        exc_type: type[BaseException],
        exc_value: BaseException,
        exc_traceback: TracebackType,
        *,
        limit: int | None = None,
        lookup_lines: bool = True,
        capture_locals: bool = False,
        compact: bool = False,
        _seen: set[int] | None = None,
    ) -> None:
        kwargs: dict[str, Any] = {}
        if sys.version_info >= (3, 10):
            kwargs["compact"] = compact

        # Capture the original exception and its cause and context as
        # TracebackExceptions
        traceback_exception_original_init(
            self,
            exc_type,
            exc_value,
            exc_traceback,
            limit=limit,
            lookup_lines=lookup_lines,
            capture_locals=capture_locals,
            _seen=_seen,
            **kwargs,
        )

        seen_was_none = _seen is None

        if _seen is None:
            _seen = set()

        # Capture each of the exceptions in the ExceptionGroup along with each of
        # their causes and contexts
        if isinstance(exc_value, BaseExceptionGroup):
            embedded = []
            for exc in exc_value.exceptions:
                if id(exc) not in _seen:
                    embedded.append(
                        PatchedTracebackException(
                            type(exc),
                            exc,
                            exc.__traceback__,
                            lookup_lines=lookup_lines,
                            capture_locals=capture_locals,
                            # copy the set of _seen exceptions so that duplicates
                            # shared between sub-exceptions are not omitted
                            _seen=None if seen_was_none else set(_seen),
                        )
                    )
            self.exceptions = embedded
            self.msg = exc_value.message
        else:
            self.exceptions = None
        self.__notes__ = getattr(exc_value, "__notes__", ())

    def format(self, *, chain=True, _ctx=None):
        if _ctx is None:
            _ctx = _ExceptionPrintContext()

        output = []
        exc = self
        if chain:
            while exc:
                if exc.__cause__ is not None:
                    chained_msg = _cause_message
                    chained_exc = exc.__cause__
                elif exc.__context__ is not None and not exc.__suppress_context__:
                    chained_msg = _context_message
                    chained_exc = exc.__context__
                else:
                    chained_msg = None
                    chained_exc = None

                output.append((chained_msg, exc))
                exc = chained_exc
        else:
            output.append((None, exc))

        for msg, exc in reversed(output):
            if msg is not None:
                yield from _ctx.emit(msg)
            if exc.exceptions is None:
                if exc.stack:
                    yield from _ctx.emit("Traceback (most recent call last):\n")
                    yield from _ctx.emit(exc.stack.format())
                yield from _ctx.emit(exc.format_exception_only())
            elif _ctx.exception_group_depth > max_group_depth:
                # exception group, but depth exceeds limit
                yield from _ctx.emit(f"... (max_group_depth is {max_group_depth})\n")
            else:
                # format exception group
                is_toplevel = _ctx.exception_group_depth == 0
                if is_toplevel:
                    _ctx.exception_group_depth += 1

                if exc.stack:
                    yield from _ctx.emit(
                        "Exception Group Traceback (most recent call last):\n",
                        margin_char="+" if is_toplevel else None,
                    )
                    yield from _ctx.emit(exc.stack.format())

                yield from _ctx.emit(exc.format_exception_only())
                num_excs = len(exc.exceptions)
                if num_excs <= max_group_width:
                    n = num_excs
                else:
                    n = max_group_width + 1
                _ctx.need_close = False
                for i in range(n):
                    last_exc = i == n - 1
                    if last_exc:
                        # The closing frame may be added by a recursive call
                        _ctx.need_close = True

                    if max_group_width is not None:
                        truncated = i >= max_group_width
                    else:
                        truncated = False
                    title = f"{i + 1}" if not truncated else "..."
                    yield (
                        _ctx.indent()
                        + ("+-" if i == 0 else "  ")
                        + f"+---------------- {title} ----------------\n"
                    )
                    _ctx.exception_group_depth += 1
                    if not truncated:
                        yield from exc.exceptions[i].format(chain=chain, _ctx=_ctx)
                    else:
                        remaining = num_excs - max_group_width
                        plural = "s" if remaining > 1 else ""
                        yield from _ctx.emit(
                            f"and {remaining} more exception{plural}\n"
                        )

                    if last_exc and _ctx.need_close:
                        yield _ctx.indent() + "+------------------------------------\n"
                        _ctx.need_close = False
                    _ctx.exception_group_depth -= 1

                if is_toplevel:
                    assert _ctx.exception_group_depth == 1
                    _ctx.exception_group_depth = 0

    def format_exception_only(self):
        """Format the exception part of the traceback.
        The return value is a generator of strings, each ending in a newline.
        Normally, the generator emits a single string; however, for
        SyntaxError exceptions, it emits several lines that (when
        printed) display detailed information about where the syntax
        error occurred.
        The message indicating which exception occurred is always the last
        string in the output.
        """
        if self.exc_type is None:
            yield traceback._format_final_exc_line(None, self._str)
            return

        stype = self.exc_type.__qualname__
        smod = self.exc_type.__module__
        if smod not in ("__main__", "builtins"):
            if not isinstance(smod, str):
                smod = "<unknown>"
            stype = smod + "." + stype

        if not issubclass(self.exc_type, SyntaxError):
            yield _format_final_exc_line(stype, self._str)
        elif traceback_exception_format_syntax_error is not None:
            yield from traceback_exception_format_syntax_error(self, stype)
        else:
            yield from traceback_exception_original_format_exception_only(self)

        if isinstance(self.__notes__, collections.abc.Sequence):
            for note in self.__notes__:
                note = _safe_string(note, "note")
                yield from [line + "\n" for line in note.split("\n")]
        elif self.__notes__ is not None:
            yield _safe_string(self.__notes__, "__notes__", func=repr)


traceback_exception_original_init = traceback.TracebackException.__init__
traceback_exception_original_format = traceback.TracebackException.format
traceback_exception_original_format_exception_only = (
    traceback.TracebackException.format_exception_only
)
traceback_exception_format_syntax_error = getattr(
    traceback.TracebackException, "_format_syntax_error", None
)
if sys.excepthook is sys.__excepthook__:
    traceback.TracebackException.__init__ = (  # type: ignore[assignment]
        PatchedTracebackException.__init__
    )
    traceback.TracebackException.format = (  # type: ignore[assignment]
        PatchedTracebackException.format
    )
    traceback.TracebackException.format_exception_only = (  # type: ignore[assignment]
        PatchedTracebackException.format_exception_only
    )
    sys.excepthook = exceptiongroup_excepthook


@singledispatch
def format_exception_only(__exc: BaseException) -> List[str]:
    return list(
        PatchedTracebackException(type(__exc), __exc, None).format_exception_only()
    )


@format_exception_only.register
def _(__exc: type, value: BaseException) -> List[str]:
    return format_exception_only(value)


@singledispatch
def format_exception(
    __exc: BaseException,
    limit: Optional[int] = None,
    chain: bool = True,
) -> List[str]:
    return list(
        PatchedTracebackException(
            type(__exc), __exc, __exc.__traceback__, limit=limit
        ).format(chain=chain)
    )


@format_exception.register
def _(
    __exc: type,
    value: BaseException,
    tb: TracebackType,
    limit: Optional[int] = None,
    chain: bool = True,
) -> List[str]:
    return format_exception(value, limit, chain)


@singledispatch
def print_exception(
    __exc: BaseException,
    limit: Optional[int] = None,
    file: Any = None,
    chain: bool = True,
) -> None:
    if file is None:
        file = sys.stderr

    for line in PatchedTracebackException(
        type(__exc), __exc, __exc.__traceback__, limit=limit
    ).format(chain=chain):
        print(line, file=file, end="")


@print_exception.register
def _(
    __exc: type,
    value: BaseException,
    tb: TracebackType,
    limit: Optional[int] = None,
    file: Any = None,
    chain: bool = True,
) -> None:
    print_exception(value, limit, file, chain)


def print_exc(
    limit: Optional[int] = None,
    file: Any | None = None,
    chain: bool = True,
) -> None:
    value = sys.exc_info()[1]
    print_exception(value, limit, file, chain)



import base64
from urllib.parse import unquote_plus, quote_plus


from padding_oracle import (
    base64_encode,
    base64_decode,
    urlencode,
    urldecode,
)


def test_base64_encode():
    plaintext = b'the quick brown fox jumps over the lazy dog'
    expected = base64.b64encode(plaintext).decode()
    assert expected == base64_encode(plaintext)
    assert expected == base64_encode(list(plaintext))
    assert expected == base64_encode(plaintext.decode())


def test_base64_decode():
    plaintext = b'the quick brown fox jumps over the lazy dog'
    encoded = base64.b64encode(plaintext).decode()
    assert plaintext == base64_decode(encoded)
    assert plaintext == base64_decode(encoded.encode())
    assert plaintext == base64_decode(list(encoded.encode()))


def test_urlencode():
    plaintext = 'the quick brown fox jumps over the lazy dog'
    assert plaintext == unquote_plus(urlencode(plaintext))
    assert plaintext == unquote_plus(urlencode(plaintext.encode()))
    assert plaintext == unquote_plus(urlencode(list(plaintext.encode())))


def test_urldecode():
    plaintext = 'the quick brown fox jumps over the lazy dog'
    encoded = quote_plus(plaintext)
    assert plaintext == urldecode(encoded)
    assert plaintext == urldecode(encoded.encode())
    assert plaintext == urldecode(list(encoded.encode()))


if __name__ == '__main__':
    test_base64_encode()
    test_base64_decode()
    test_urlencode()
    test_urldecode()



'''
Copyright (c) 2022 Yuankui Li

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
'''

import logging
import traceback
from typing import List, Union

from .encoding import to_bytes
from .solve import (
    solve, Fail, OracleFunc, ResultType,
    convert_to_bytes, remove_padding)

__all__ = [
    'padding_oracle',
]


def padding_oracle(ciphertext: Union[bytes, str],
                   block_size: int,
                   oracle: OracleFunc,
                   num_threads: int = 1,
                   log_level: int = logging.INFO,
                   null_byte: bytes = b' ',
                   return_raw: bool = False,
                   ) -> Union[bytes, List[int]]:
    '''
    Run padding oracle attack to decrypt ciphertext given a function to check
    wether the ciphertext can be decrypted successfully.

    Args:
        cipher         (bytes|str) the cipher you want to decrypt
        block_size     (int)       block size (the cipher length should be
                                   multiple of this)
        oracle         (function)  a function: oracle(cipher: bytes) -> bool
        num_threads    (int)       how many oracle functions will be run in
                                   parallel (default: 1)
        log_level      (int)       log level (default: logging.INFO)
        null_byte      (bytes|str) the default byte when plaintext are not
                                   set (default: None)
        return_raw     (bool)      do not convert plaintext into bytes and
                                   unpad (default: False)

    Returns:
        plaintext (bytes|List[int]) the decrypted plaintext
    '''

    # Check args
    if not callable(oracle):
        raise TypeError('the oracle function should be callable')
    if not isinstance(ciphertext, (bytes, str)):
        raise TypeError('cipher should have type bytes')
    if not isinstance(block_size, int):
        raise TypeError('block_size should have type int')
    if not len(ciphertext) % block_size == 0:
        raise ValueError('cipher length should be multiple of block size')
    if not 1 <= num_threads <= 1000:
        raise ValueError('num_threads should be in [1, 1000]')
    if not isinstance(null_byte, (bytes, str)):
        raise TypeError('expect null with type bytes or str')
    if not len(null_byte) == 1:
        raise ValueError('null byte should have length of 1')

    logger = get_logger()
    logger.setLevel(log_level)

    ciphertext = to_bytes(ciphertext)
    null_byte = to_bytes(null_byte)

    # Wrapper to handle exceptions from the oracle function
    def wrapped_oracle(ciphertext: bytes):
        try:
            return oracle(ciphertext)
        except Exception as e:
            logger.error(f'error in oracle with {ciphertext!r}, {e}')
            logger.debug('error details: {}'.format(traceback.format_exc()))
        return False

    def result_callback(result: ResultType):
        if isinstance(result, Fail):
            if result.is_critical:
                logger.critical(result.message)
            else:
                logger.error(result.message)

    def plaintext_callback(plaintext: bytes):
        plaintext = convert_to_bytes(plaintext, null_byte)
        logger.info(f'plaintext: {plaintext}')

    plaintext = solve(ciphertext, block_size, wrapped_oracle, num_threads,
                      result_callback, plaintext_callback)

    if not return_raw:
        plaintext = convert_to_bytes(plaintext, null_byte)
        plaintext = remove_padding(plaintext)

    return plaintext


def get_logger():
    logger = logging.getLogger('padding_oracle')
    formatter = logging.Formatter('[%(asctime)s][%(levelname)s] %(message)s')
    handler = logging.StreamHandler()
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    return logger



from padding_oracle import padding_oracle
from .cryptor import VulnerableCryptor


def test_padding_oracle_basic():
    cryptor = VulnerableCryptor()

    plaintext = b'the quick brown fox jumps over the lazy dog'
    ciphertext = cryptor.encrypt(plaintext)

    decrypted = padding_oracle(ciphertext, cryptor.block_size,
                               cryptor.oracle, 4, null_byte=b'?')
    print(decrypted)

    assert decrypted == plaintext


if __name__ == '__main__':
    test_padding_oracle_basic()



from concurrent.futures import ProcessPoolExecutor
import os
import multiprocessing as mp

from padding_oracle import remove_padding
from padding_oracle.solve import solve
from .cryptor import VulnerableCryptor

VERBOSE = False


def _test_solve(data_size):
    crypter = VulnerableCryptor()

    plaintext = os.urandom(data_size)
    ciphertext = crypter.encrypt(plaintext)

    decrypted = solve(list(ciphertext), crypter.block_size, crypter.oracle)
    decrypted = remove_padding(decrypted)

    assert plaintext == decrypted, 'decryption failed'


def test_solve_basic():
    sizes = [0, 1, 15, 16, 17]
    for size in sizes:
        if VERBOSE:
            print('test_solve_basic', size)
        _test_solve(size)


def test_solve_many_sizes():
    with ProcessPoolExecutor(mp.cpu_count()) as executor:
        futures = []
        for size in [*range(0, 64, 7), *range(64, 128, 13)]:
            futures.append((size, executor.submit(_test_solve, size)))
        for size, future in futures:
            if VERBOSE:
                print('test_solve_many_sizes', size)
            future.result()


def test_solve_random_15():
    size = 15
    with ProcessPoolExecutor(mp.cpu_count()) as executor:
        futures = [executor.submit(_test_solve, size) for _ in range(64)]
        for future in futures:
            if VERBOSE:
                print('test_solve_random_15', size)
            future.result()


if __name__ == '__main__':
    VERBOSE = True
    test_solve_basic()
    test_solve_many_sizes()
    test_solve_random_15()



import os
from .cryptor import VulnerableCryptor

VERBOSE = False


def test_cryptor():
    for size in range(1, 128):
        cryptor = VulnerableCryptor()
        plaintext = os.urandom(size)
        if VERBOSE:
            print('test_cryptor', size, plaintext)
        ciphertext = cryptor.encrypt(plaintext)
        assert len(ciphertext) % cryptor.block_size == 0
        assert cryptor.decrypt(ciphertext) == plaintext, \
            f'{len(plaintext)} {len(ciphertext)}'


if __name__ == '__main__':
    VERBOSE = True
    test_cryptor()



'''
Copyright (c) 2022 Yuankui Li

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "PIU PIU", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
'''

import asyncio
from concurrent.futures import ThreadPoolExecutor
from typing import Awaitable, Callable, NamedTuple, Set, Union, List

from .encoding import to_bytes


__all__ = [
    'solve',
    'convert_to_bytes',
    'remove_padding',
]


class Pass(NamedTuple):
    block_index: int
    index: int
    byte: int


class Fail(NamedTuple):
    block_index: int
    message: str
    is_critical: bool = False


class Done(NamedTuple):
    block_index: int
    C0: List[int]
    X1: List[int]


ResultType = Union[Pass, Fail, Done]

OracleFunc = Callable[[bytes], bool]
ResultCallback = Callable[[ResultType], bool]
PlainTextCallback = Callable[[List[int]], bool]


class Context(NamedTuple):
    block_size: int
    oracle: OracleFunc

    executor: ThreadPoolExecutor
    loop: asyncio.AbstractEventLoop

    tasks: Set[Awaitable[ResultType]]

    latest_plaintext: List[int]
    plaintext: List[int]

    result_callback: ResultCallback
    plaintext_callback: PlainTextCallback


def dummy_callback(*a, **ka):
    pass


def solve(ciphertext: bytes,
          block_size: int,
          oracle: OracleFunc,
          parallel: int = 1,
          result_callback: ResultCallback = dummy_callback,
          plaintext_callback: PlainTextCallback = dummy_callback,
          ) -> List[int]:

    loop = asyncio.new_event_loop()
    future = solve_async(ciphertext, block_size, oracle, parallel,
                         result_callback, plaintext_callback)
    return loop.run_until_complete(future)


async def solve_async(ciphertext: bytes,
                      block_size: int,
                      oracle: OracleFunc,
                      parallel: int = 1,
                      result_callback: ResultCallback = dummy_callback,
                      plaintext_callback: PlainTextCallback = dummy_callback,
                      ) -> List[int]:

    ciphertext = list(ciphertext)

    if not len(ciphertext) % block_size == 0:
        raise ValueError('ciphertext length must be a multiple of block_size')
    if not len(ciphertext) // block_size > 1:
        raise ValueError('cannot solve with only one block')

    ctx = create_solve_context(ciphertext, block_size, oracle, parallel,
                               result_callback, plaintext_callback)

    while True:
        done_tasks, _ = await asyncio.wait(ctx.tasks,
                                           return_when=asyncio.FIRST_COMPLETED)

        for task in done_tasks:
            result = await task

            ctx.result_callback(result)
            ctx.tasks.remove(task)

            if isinstance(result, Pass):
                update_latest_plaintext(
                    ctx, result.block_index, result.index, result.byte)
            if isinstance(result, Done):
                update_plaintext(ctx, result.block_index, result.C0, result.X1)

        if len(ctx.tasks) == 0:
            break

    # Check if any block failed
    error_block_indices = set()

    for i, byte in enumerate(ctx.plaintext):
        if byte is None:
            error_block_indices.add(i // block_size + 1)

    for idx in error_block_indices:
        result_callback(Fail(idx, f'cannot decrypt cipher block {idx}', True))

    return ctx.plaintext


def create_solve_context(ciphertext, block_size, oracle, parallel,
                         result_callback, plaintext_callback) -> Context:
    tasks = set()

    cipher_blocks = []
    for i in range(0, len(ciphertext), block_size):
        cipher_blocks.append(ciphertext[i:i+block_size])

    plaintext = [None] * (len(cipher_blocks) - 1) * block_size
    latest_plaintext = plaintext.copy()

    executor = ThreadPoolExecutor(parallel)
    loop = asyncio.get_event_loop()
    ctx = Context(block_size, oracle, executor, loop, tasks,
                  latest_plaintext, plaintext,
                  result_callback, plaintext_callback)

    for i in range(1, len(cipher_blocks)):
        run_block_task(ctx, i, cipher_blocks[i-1], cipher_blocks[i], [])

    return ctx


def run_block_task(ctx: Context, block_index, C0, C1, X1):
    future = solve_block(ctx, block_index, C0, C1, X1)
    task = ctx.loop.create_task(future)
    ctx.tasks.add(task)


async def solve_block(ctx: Context, block_index: int, C0: List[int],
                      C1: List[int], X1: List[int] = []) -> ResultType:
    # X1 = decrypt(C1)
    # P1 = xor(C0, X1)

    if len(X1) == ctx.block_size:
        return Done(block_index, C0, X1)

    assert len(C0) == ctx.block_size
    assert len(C1) == ctx.block_size
    assert len(X1) in range(ctx.block_size)

    index = ctx.block_size - len(X1) - 1
    padding = len(X1) + 1

    C0_test = C0.copy()
    for i in range(len(X1)):
        C0_test[-i-1] = X1[-i-1] ^ padding
    hits = list(await get_oracle_hits(ctx, C0_test, C1, index))

    invalid = len(X1) == 0 and len(hits) not in (1, 2)
    invalid |= len(X1) > 0 and len(hits) != 1
    if invalid:
        message = 'unexpected number of hits: block={} index={} n={}' \
            .format(block_index, index, len(hits))
        return Fail(block_index, message)

    for byte in hits:
        X1_test = [byte ^ padding, *X1]
        run_block_task(ctx, block_index, C0, C1, X1_test)

    return Pass(block_index, index, byte ^ padding ^ C0[index])


async def get_oracle_hits(ctx: Context, C0: List[int], C1: List[int],
                          index: int):

    C0 = C0.copy()
    futures = {}

    for byte in range(256):
        C0[index] = byte
        ciphertext = bytes(C0 + C1)
        futures[byte] = ctx.loop.run_in_executor(
            ctx.executor, ctx.oracle, ciphertext)

    hits = []

    for byte, future in futures.items():
        is_valid = await future
        if is_valid:
            hits.append(byte)

    return hits


def update_latest_plaintext(ctx: Context, block_index: int, index: int,
                            byte: int):

    i = (block_index - 1) * ctx.block_size + index
    ctx.latest_plaintext[i] = byte
    ctx.plaintext_callback(ctx.latest_plaintext)


def update_plaintext(ctx: Context, block_index: int, C0: List[int],
                     X1: List[int]):

    assert len(C0) == len(X1) == ctx.block_size
    block = compute_plaintext(C0, X1)

    i = (block_index - 1) * ctx.block_size
    ctx.latest_plaintext[i:i+ctx.block_size] = block
    ctx.plaintext[i:i+ctx.block_size] = block
    ctx.plaintext_callback(ctx.plaintext)


def compute_plaintext(C0: List[int], X1: List[int]):
    return [c ^ x for c, x in zip(C0, X1)]


def convert_to_bytes(byte_list: List[int], replacement=b' '):
    '''
    Convert a list of int into bytes, replace invalid byte with replacement.
    '''
    for i, byte in enumerate(list(byte_list)):
        if isinstance(byte, int) and byte in range(256):
            pass
        elif isinstance(byte, bytes):
            byte = ord(byte)
        else:
            byte = ord(replacement)
        byte_list[i] = byte
    return bytes(byte_list)


def remove_padding(data: Union[str, bytes, List[int]]) -> bytes:
    '''
    Remove PKCS#7 padding bytes.
    '''
    data = to_bytes(data)
    return data[:-data[-1]]
